##
# Copyright 2022 Charles Y. Choi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse


class CommandLineParser:
    def __init__(self):

        description = """
        %(prog)s - generate SVG diagrams of control register-style data formats
        """
        epilog = """
        Input file details can be found at <https://github.com/kickingvegas/regfmt>. 
        Please report issues to <https://github.com/kickingvegas/regfmt/issues>. 
        """
                
        self.parser = argparse.ArgumentParser(description=description,
                                              epilog=epilog)
        self.parser.add_argument('-v', '--version',
                                 action='store_true',
                                 help='print version information and exit')
        
        self.parser.add_argument('-o', '--output',
                                 action='store',
                                 default='-',
                                 help="output file (default: '-' for stdout)")

        self.parser.add_argument('-s', '--style',
                                 action='store',
                                 default=None,
                                 help='CSS style file')

        self.parser.add_argument('input', nargs='?', default='input.yaml',
                                 help='input register format YAML file (default: input.yaml)')
                
    def run(self):
        return self.parser.parse_args()



#!/usr/bin/env python3
##
# Copyright 2022 Charles Y. Choi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
from regfmt import VERSION
from regfmt import CommandLineParser
from regfmt import InputLoadAndValidate
from regfmt import TopLevel
from regfmt import DRCChecker
from regfmt import SVGWriter


class RegisterFormat:
    def __init__(self, parsedArguments):
        self.version = VERSION
        self.stdout = sys.stdout
        self.stdin = sys.stdin
        self.stderr = sys.stderr
        self.parsedArguments = parsedArguments

        if parsedArguments.version:
            sys.stdout.write('{0}\n'.format(self.version))
            sys.exit(0)

        if parsedArguments.output != '-':
            outfile = open('{0}'.format(parsedArguments.output), 'w')
            self.stdout = outfile

        if not os.path.exists(parsedArguments.input):
            message = ('ERROR: file "{0}" does not exist. '
                       'Please specify an input file.\n')
            sys.stderr.write(message.format(parsedArguments.input))
            sys.exit(1)

    def run(self):
        # Load and validate input YAML
        loader = InputLoadAndValidate(self.parsedArguments)
        inputYAML = loader.loadAndValidate()
        # print(inputYAML)

        # Deserialize input YAML into native object DB
        registerDB = TopLevel(config=inputYAML)

        # DRC Check native object DB
        drcChecker = DRCChecker()
        drcChecker.check(registerDB)
        drcChecker.subIndexFields(registerDB)

        # Render SVG
        svgWriter = SVGWriter(registerDB,
                              self.stdout,
                              configFileName=self.parsedArguments.style)
        svgWriter.writeSVG()

        # wrap up
        if self.stdout != sys.stdout:
            self.stdout.close()


if __name__ == '__main__':
    app = RegisterFormat(CommandLineParser().run())
    app.run()



##
# Copyright 2022 Charles Y. Choi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from regfmtlib import Endian

class TestEndian(unittest.TestCase):
    def test_littleByte(self):
        self.assertEqual(Endian.littleByte.value, "littleByte")

    def test_bigByte(self):
        self.assertEqual(Endian.bigByte.value, "bigByte")
        
    def test_littleBit(self):
        self.assertEqual(Endian.littleBit.value, "littleBit")

    def test_bigBit(self):
        self.assertEqual(Endian.bigBit.value, "bigBit")

if __name__ == '__main__':
    unittest.main()
    



##
# Copyright 2022 Charles Y. Choi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from PIL import ImageFont
import sys
from regfmt import BASE_FONT_NAME

class StrokeLinecap(Enum):
    butt = 'butt'
    round = 'round'
    square = 'square'


class FontStyle(Enum):
    normal = 'normal'
    italic = 'italic'
    oblique = 'oblique'


class FontWeight(Enum):
    normal = 'normal'
    bold = 'bold'
    bolder = 'bolder'
    lighter = 'lighter'


class TextStyle:
    def __init__(self,
                 fontFamily: [str] = None,
                 fontSize: str = None,
                 fontStyle: FontStyle = None,
                 fontWeight: FontWeight = None,
                 fill: str = None):
        self.fontFamily = fontFamily
        self.fontSize = fontSize
        self.fontStyle = fontStyle
        self.fontWeight = fontWeight
        self.fill = fill
        self.font = None

    def truetype(self):
        if self.font:
            return self.font

        baseFontSize = self.fontSize

        if 'pt' in baseFontSize:
            baseFontSize = int(round(float(baseFontSize.replace('pt', ''))))

        else:
            try:
                baseFontSize = int(float(round(baseFontSize)))
            except:
                message = ('WARNING: body font-size specification of "{}" is unsupported in '
                           'CSS file for sizing the geometry of register fields. '
                           'Coercing font size value to {}pt.\n')
                sys.stderr.write(message.format(fontSize, 12))
                baseFontSize = 12

        for fontName in self.fontFamily:
            try:
                font = ImageFont.truetype(fontName, baseFontSize)
                continue
            except OSError:
                
                font = ImageFont.truetype(BASE_FONT_NAME, baseFontSize)
                break

        return font

class LineStyle:
    def __init__(self,
                 stroke: str = 'grey',
                 strokeWidth: str = '0.8',
                 strokeLinecap: StrokeLinecap = StrokeLinecap.butt
                 ):
        self.stroke: str = stroke
        self.strokeWidth: str = strokeWidth
        self.strokeLinecap: StrokeLinecap = strokeLinecap


class RectStyle:
    def __init__(self,
                 fill: str = None,
                 stroke: str = None,
                 strokeWidth: str = None,
                 strokeLinecap: StrokeLinecap = None
                 ):
        # TODO: support fill opacity
        # CSS color type
        self.fill: str = fill
        # CSS color type
        self.stroke: str = stroke
        self.strokeWidth: str = strokeWidth
        self.strokeLinecap: StrokeLinecap = strokeLinecap


class BaseStyle(TextStyle, RectStyle):
    def __init__(self):
        TextStyle.__init__(self,
                           fontFamily=['Futura', 'Helvetica', 'Arial', 'sans-serif'],
                           fontSize='12pt',
                           fontStyle=FontStyle.normal,
                           fontWeight=FontWeight.normal,
                           fill='black')

        RectStyle.__init__(self,
                           fill='none',
                           stroke='black',
                           strokeWidth='0.5',
                           strokeLinecap=StrokeLinecap.butt)


class StyleSheet:
    def __init__(self):
        self.body = BaseStyle()
        self.register = RectStyle()
        self.field = RectStyle()
        self.registerName = TextStyle()
        self.fieldName = TextStyle()
        self.fieldIndex = TextStyle(fontSize='10pt')
        self.fieldNameLine = LineStyle()



##
# Copyright 2022 Charles Y. Choi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
from regfmt.cssstyles import *
from xml.dom.minidom import parseString
from PIL import ImageFont
from collections import UserList
from regfmt import BASE_FONT_NAME

BASE_FONT_SIZE = 12

def bboxWidth(bbox):
    return (bbox[2] - bbox[0])

def bboxHeight(bbox):
    return (bbox[3] - bbox[1])

def matrixMult(X, Y):
    result = [[sum(a*b for a,b in zip(X_row,Y_col)) for Y_col in zip(*Y)] for X_row in X]
    return result

def translateTransform(x, y):
    result = [[1.0, 0.0, x],
              [0.0, 1.0, y],
              [0.0, 0.0, 1.0]]
    return result

def vector2D(x, y):
    result = [[x], [y], [1.0]]
    return result

def coordinateFromVector2D(V):
    result = (V[0][0], V[1][0])
    return result

def numToUnitString(num, unit: str=''):
    if unit == '':
        result = '{0}'.format(num)
    else:
        result = '{0}{1}'.format(num, unit)
    return result

def cssFontToImageFont(fontFamily, fontSize):
    result = None
    for baseFontname in fontFamily:
        baseFontSize = fontSize

        # !!!: only pt is supported for geometry calculations

        if 'pt' in baseFontSize:
            baseFontSize = int(round(float(baseFontSize.replace('pt', ''))))
        else:
            try:
                baseFontSize = int(float(round(baseFontSize)))
            except:
                message = ('WARNING: body font-size specification of "{}" is unsupported in '
                           'CSS file for sizing the geometry of register fields. '
                           'Coercing font size value to {}pt.\n')
                sys.stderr.write(message.format(fontSize, BASE_FONT_SIZE))
                baseFontSize = BASE_FONT_SIZE

        try:
            result = ImageFont.truetype(baseFontname, baseFontSize)
            break
        except OSError:
            continue

    if result is None:
        result = ImageFont.truetype(BASE_FONT_NAME, baseFontSize)

    return result

def getTextFrame(text, font, anchor='ls', baseDPI=96.0):
    """
    use baseline by default

    scalingFactor adjusts for the screen resolution which presumes 96 DPI.

    :param text:
    :param font:
    :param anchor:
    :return:
    """
    scalingFactor = baseDPI / 64.0
    bbox64 = font.getbbox(text=text, anchor=anchor)
    bbox = tuple(map(lambda x: x * scalingFactor, bbox64))
    length = font.getlength(text=text)
    frame = Frame(x=bbox[0],
                  y=(bbox[1] * -1),
                  width=bboxWidth(bbox),
                  height=bboxHeight(bbox))
    return frame

def getBitFieldSize(font):
    emFrame = getTextFrame(text='M', font=font)
    emFrame.size.width *= 2
    emFrame.size.height *= 4
    return emFrame.size


class Group(UserList):
    def writeDOM(self, children, doc):
        """
        Recursive routine to populate children with all DOM elements corresponding
        to the nodes in this instance of Group.

        :param children: list of DOM elements
        :param doc: SVG XML document
        :return: none
        """
        for element in self.data:
            if isinstance(element, Group):
                element.writeDOM(children, doc)
            else:
                shape: Shape = element
                children.append(shape.writeDOM(doc))

    def translate(self, dx: float=0.0, dy: float=0.0):
        for element in self.data:
            if isinstance(element, Group):
                element.translate(dx, dy)
            else:
                #print('translating {0} by {1}, {2}'.format(element, dx, dy))
                shape: Shape = element
                shape.translate(dx, dy)

class Point:
    def __init__(self, x: float=0.0, y: float=0.0):
        self.x: float = x
        self.y: float = y

class Size:
    def __init__(self, width: float=0.0, height: float=0.0):
        self.width: float = width
        self.height: float = height

class Frame:
    def __init__(self, x: float = 0.0, y: float = 0.0, width: float = 0.0, height: float = 0.0):
        self.origin: Point = Point(x, y)
        self.size: Size = Size(width, height)

class Geometry:
    def writeDOM(self, doc):
        # virtual method; intended to be overridden
        pass

    def translate(self, dx: float=0.0, dy: float=0.0):
        # virutal method; indented to be overwritten
        pass

class Shape(Geometry):
    def __init__(self, x: float = 0.0, y: float = 0.0, width: float = 0.0, height: float = 0.0):
        Geometry.__init__(self)
        self.frame: Frame = Frame(x, y, width, height)

    def origin(self):
        return self.frame.origin

    def size(self):
        return self.frame.size

    def setOrigin(self, x: float=0.0, y: float=0.0):
        self.frame.origin.x = x
        self.frame.origin.y = y

    def setSize(self, width: float=0.0, height: float=0.0):
        self.frame.size.width = width
        self.frame.size.height = height

    def translate(self, dx: float = 0.0, dy: float = 0.0):
        T = translateTransform(dx, dy)
        V = vector2D(self.origin().x, self.origin().y)
        transformedV = matrixMult(T, V)
        newX, newY = coordinateFromVector2D(transformedV)
        self.setOrigin(newX, newY)


class Line(Geometry):
    def __init__(self,
                 x1: float = 0.0,
                 y1: float = 0.0,
                 x2: float = 0.0,
                 y2: float = 0.0,
                 style: LineStyle = LineStyle()):
        Geometry.__init__(self)
        self.x1 = x1
        self.y1 = y1
        self.x2 = x2
        self.y2 = y2
        self.style: LineStyle = style

    def writeDOM(self, doc):
        lineElement = doc.createElement('line')
        lineElement.setAttribute('x1', str(self.x1))
        lineElement.setAttribute('y1', str(self.y1))
        lineElement.setAttribute('x2', str(self.x2))
        lineElement.setAttribute('y2', str(self.y2))
        lineElement.setAttribute('stroke', self.style.stroke)
        lineElement.setAttribute('stroke-width', self.style.strokeWidth)
        lineElement.setAttribute('stroke-linecap', self.style.strokeLinecap.value)
        return lineElement

    def translate(self, dx: float=0.0, dy: float=0.0):
        T = translateTransform(dx, dy)
        V = vector2D(self.x1, self.y1)
        transformedV = matrixMult(T, V)
        newX, newY = coordinateFromVector2D(transformedV)
        self.x1 = newX
        self.y1 = newY

        V = vector2D(self.x2, self.y2)
        transformedV = matrixMult(T, V)
        newX, newY = coordinateFromVector2D(transformedV)
        self.x2 = newX
        self.y2 = newY


class Rect(Shape):
    def __init__(self,
                 x: float = 0.0, y: float = 0.0, width: float = 0.0, height: float = 0.0,
                 style: RectStyle = RectStyle(),
                 ):
        Shape.__init__(self, x, y, width, height)
        self.style: RectStyle = style

    def writeDOM(self, doc):
        rectElement = doc.createElement('rect')
        rectElement.setAttribute('x', str(self.frame.origin.x))
        rectElement.setAttribute('y', str(self.frame.origin.y))
        rectElement.setAttribute('width', str(self.frame.size.width))
        rectElement.setAttribute('height', str(self.frame.size.height))
        rectElement.setAttribute('fill', self.style.fill)
        rectElement.setAttribute('stroke', self.style.stroke)
        rectElement.setAttribute('stroke-width', self.style.strokeWidth)
        rectElement.setAttribute('stroke-linecap', self.style.strokeLinecap.value)
        return rectElement

class Text(Shape):
    def __init__(self,
                 text="something",
                 x: float = 0.0, y: float = 0.0, width: float = 0.0, height: float = 0.0,
                 textAnchor: str='start',
                 style: TextStyle = TextStyle()
                 ):

        Shape.__init__(self, x, y, width, height)
        self.text = text
        self.textAnchor = textAnchor
        self.style = style

    def writeDOM(self, doc):
        textElement = doc.createElement('text')
        textElement.setAttribute('x', str(self.frame.origin.x))
        textElement.setAttribute('y', str(self.frame.origin.y))
        textElement.setAttribute('font-size', self.style.fontSize)
        textElement.setAttribute('font-family', ','.join(self.style.fontFamily))
        textElement.setAttribute('font-style', self.style.fontStyle.value)
        textElement.setAttribute('font-weight', self.style.fontWeight.value)
        textElement.setAttribute('text-anchor', self.textAnchor)
        textElement.setAttribute('fill', self.style.fill)

        if '<sub>' in self.text:
            # don't think this is going to work.
            newText = '<p>{}</p>'.format(self.text)
            xmlFragment = parseString(newText)
            textNode = xmlFragment.documentElement

        else:
            textNode = doc.createTextNode(self.text)
        textElement.appendChild(textNode)
        return textElement


if __name__ == '__main__':
    
    group = Group()
    group.append(Rect((0, 0, 10, 10)))
    group.append(Rect((0, 5, 29, 12)))
    group.append(Text())
    
    topGroup = Group()
    
    topGroup.append(group)
    


        



##
# Copyright 2022 Charles Y. Choi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

class Field:
    def __init__(self, config=None):
        self.name: str = None
        self.width: int = None
        self.leftIndex: int = None
        self.rightIndex: int = None

        # !!!: Note config should already be validated.
        if config:
            if 'name' in config:
                self.name = config['name']
            if 'width' in config:
                self.width = config['width']



##
# Copyright 2022 Charles Y. Choi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from xml.dom.minidom import getDOMImplementation
from regfmt import TopLevel, FieldNameAlign, VERSION
from regfmt import Register
from regfmt.svggeometry import *
from regfmt.cssparser import parseCSS, cascadeStyles
from regfmt.centeralignlayout import writeFieldNameCenterSVG
from regfmt.stairleftlayout import writeStairLeftSVG
from tinycss2.ast import Declaration

class SVGWriter:
    def __init__(self, registerDB: TopLevel, outfile, configFileName=None):
        self.registerDB = registerDB
        self.outfile = outfile
        self.styleSheet = StyleSheet()
        try:
            parseCSS(configFileName=configFileName, styleSheet=self.styleSheet)
        except FileNotFoundError as err:
            message = 'ERROR: {}: "{}"  Exiting…\n'.format(err.strerror, err.filename)
            sys.stderr.write(message)
            sys.exit(err.errno)

        except ValueError as err:
            message = 'ERROR: {} Exiting…\n'.format(err.args[1])
            if isinstance(err.args[0], Declaration):
                declaration: Declaration = err.args[0]
                message = 'ERROR: {} ({}, {}): {}'.format(configFileName,
                                                          declaration.source_line,
                                                          declaration.source_column,
                                                          err.args[1])
            sys.stderr.write(message)
            sys.exit(1)


        cascadeStyles(self.styleSheet)
        self.font = cssFontToImageFont(fontFamily=self.styleSheet.body.fontFamily,
                                       fontSize=self.styleSheet.body.fontSize)

    def createDocument(self, origin=None, size=None):
        dom = getDOMImplementation()
        doc = dom.createDocument(None, 'svg', None)
        topElement = doc.documentElement
        topElement.setAttribute('version', '1.1')
        topElement.setAttribute('xmlns', "http://www.w3.org/2000/svg")
        watermark = 'diagram generated by regfmt {} https://github.com/kickingvegas/regfmt'.format(VERSION)
        topElement.appendChild(doc.createComment(watermark))
        if origin:
            topElement.setAttribute('x', str(origin.x))
            topElement.setAttribute('y', str(origin.y))

        if size:
            topElement.setAttribute('width', str(size.width))
            topElement.setAttribute('height', str(size.height))

        return doc, topElement


    def compositeSVG(self, doc, topElement, topGroup, outfile):
        children = []
        topGroup.writeDOM(children, doc)
        for child in children:
            topElement.appendChild(child)
        doc.writexml(outfile, encoding='utf-8', addindent="  ", newl="\n")

    def writeSVG(self, registerDB: TopLevel = None):
        if registerDB is None:
            registerDB = self.registerDB

        # Layout Geometry - all decisions on the graphical layout of a register instance are done here.

        # Initialize container for geometry
        topGroup = Group()
        if registerDB.layout.fieldNameAlign == FieldNameAlign.center:
            origin, size = writeFieldNameCenterSVG(registerDB=registerDB,
                                                   topGroup=topGroup,
                                                   styleSheet=self.styleSheet,
                                                   font=self.font)

        elif registerDB.layout.fieldNameAlign == FieldNameAlign.stairLeft:
            origin, size = writeStairLeftSVG(registerDB=registerDB,
                                             topGroup=topGroup,
                                             styleSheet=self.styleSheet,
                                             font=self.font)
            #origin, size = self.writeFieldNameStairLeftSVG(registerDB, topGroup)

        doc, topElement = self.createDocument(origin=origin, size=size)

        # Write DOM and composite to SVG
        self.compositeSVG(doc, topElement, topGroup, self.outfile)

    def writeFieldNameStairLeftSVG(self, registerDB, topGroup):
        registers: [Register] = registerDB.registers
        bitFieldSize = self.getBitFieldSize(self.font)
        # Generate groups, each group holding a single register
        # layoutGeometry, transformGeometry, writeDOM, compositeSVG
        for register in registers:
            # define untransformed geometry for register
            registerHeight = bitFieldSize.height
            registerWidth = bitFieldSize.width * register.width

            registerGroup = Group()
            topGroup.append(registerGroup)

            if register.name is not None:
                registerNameBbox = self.font.getbbox(register.name, anchor='lt')
                registerNameHeight = bboxHeight(registerNameBbox)
                registerNameWidth = bboxWidth(registerNameBbox)
                registerGroup.append(Text(register.name,
                                          x=registerWidth + 10,
                                          y=registerNameHeight + (registerHeight / 2.0) - (registerNameHeight / 2.0),
                                          style=self.styleSheet.registerName
                                          ))

            displacementX = 0
            fieldDisplacementY = 0

            fieldNames = list(filter(lambda x: x is not None, [field.name for field in register.fields]))
            fieldCount = len(fieldNames)
            fieldNameHeight = 0

            for name in fieldNames:
                fieldNameBbox = self.font.getbbox(name, anchor='la')
                fieldNameHeight = bboxHeight(fieldNameBbox)
                break

            fieldDisplacementY = (fieldNameHeight + 10.0) * (fieldCount)

            for field in register.fields:
                fieldWidth = field.width * bitFieldSize.width
                fieldHeight = registerHeight

                fieldX = 0
                fieldY = 0

                T = translateTransform(displacementX, 0.0)
                V = vector2D(fieldX, fieldY)
                transformedV = matrixMult(T, V)
                fieldX, fieldY = coordinateFromVector2D(transformedV)

                registerGroup.append(Rect(x=fieldX,
                                          y=fieldY,
                                          width=fieldWidth,
                                          height=fieldHeight,
                                          style=self.styleSheet.field))

                # TODO: support stair layout

                fieldNameBbox = self.font.getbbox(field.name, anchor='la')
                fieldNameHeight = bboxHeight(fieldNameBbox)
                fieldNameWidth = bboxWidth(fieldNameBbox)

                # print(fieldDisplacementY)
                registerGroup.append(Text(field.name,
                                          x=fieldNameBbox[0] + (fieldWidth / 2.0) + displacementX + 3,
                                          y=fieldNameBbox[3] + fieldDisplacementY + fieldNameHeight + 10,
                                          textAnchor='left',
                                          style=self.styleSheet.fieldName
                                          ))
                fieldDisplacementY -= (fieldNameHeight + 10.0)

                # Draw Line
                registerGroup.append(Line(x1=displacementX + (fieldWidth / 2.0),
                                          y1=registerHeight,
                                          x2=displacementX + (fieldWidth / 2.0),
                                          y2=fieldNameBbox[3] + fieldDisplacementY + fieldNameHeight * 2 + 15,
                                          style=self.styleSheet.fieldNameLine))

                if field.width > 1:
                    registerGroup.append(Text(str(field.leftIndex),
                                              x=fieldX + 3,
                                              y=fieldY + fieldHeight - 3,
                                              textAnchor='start',
                                              style=self.styleSheet.fieldIndex
                                              ))

                    registerGroup.append(Text(str(field.rightIndex),
                                              x=fieldX + fieldWidth - 3,
                                              y=fieldY + fieldHeight - 3,
                                              textAnchor='end',
                                              style=self.styleSheet.fieldIndex
                                              ))
                else:
                    registerGroup.append(Text(str(field.leftIndex),
                                              x=fieldX + (fieldWidth / 2.0),
                                              y=fieldY + fieldHeight - 3,
                                              textAnchor='middle',
                                              style=self.styleSheet.fieldIndex
                                              ))

                displacementX += fieldWidth

            # break
        # Transform Geometry

        # TODO: need to fix stairLeft for multiple registers

        displacementY = 0
        for group in topGroup:
            group.translate(0, displacementY)
            displacementY += (fieldNameHeight + 10.0) * (fieldCount)

        # TODO: implement this for real.
        #origin = Point()
        #size = Size(width=registerWidth + registerNameWidth + 20,
        #height=registerHeight)

        return (None, None)



##
# Copyright 2022 Charles Y. Choi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import errno
import tinycss2
from regfmt.cssstyles import *

classNameMap = {
                'register-name': 'registerName',
                'field-name': 'fieldName',
                'field-index': 'fieldIndex',
                'field-name-line': 'fieldNameLine'
                }

def parseCSS(configFileName: str, styleSheet: StyleSheet):
    if configFileName is None:
        return

    if not os.path.exists(configFileName):
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), configFileName)

    css = None
    with open(configFileName, 'r') as configFile:
        css = configFile.read()

    if css == '':
        raise ValueError('CSS file "{}" is empty.'.format(configFileName))

    rules = tinycss2.parse_stylesheet(css, skip_comments=True, skip_whitespace=True)
    for rule in rules:
        # Extract declarations
        declarations = tinycss2.parse_declaration_list(rule.content,
                                                       skip_comments=True,
                                                       skip_whitespace=True)

        filterDeclarationErrors(configFileName, declarations)

        className = rule.prelude[0].value
        #print(className)

        if className not in ('body', 'register', 'field', 'register-name',
                             'field-name', 'field-index', 'field-name-line'):
            # list of legal names
            continue

        elif className == 'body':
            legalAttributes = ('font-family', 'font-size', 'font-style', 'font-weight',
                               'fill',
                               'stroke', 'stroke-width', 'stroke-linecap')

            try:
                for declaration in declarations:
                    if declaration.name not in legalAttributes:
                        continue

                    elif declaration.name == 'font-family':
                        styleSheet.body.fontFamily = extractFontFamily(declaration)

                    elif declaration.name == 'font-size':
                        styleSheet.body.fontSize = extractDimensionalValue(declaration)
                        #styleSheet.registerName.fontSize = styleSheet.body.fontSize
                        #styleSheet.fieldName.fontSize = styleSheet.body.fontSize
                        # TODO: deal with fieldIndex

                    elif declaration.name == 'font-style':
                        styleSheet.body.fontStyle = extractEnum(declaration, FontStyle)

                    elif declaration.name == 'font-weight':
                        styleSheet.body.fontWeight = extractEnum(declaration, FontWeight)

                    elif declaration.name == 'fill':
                        styleSheet.body.fill = extractColor(declaration)

                    elif declaration.name == 'stroke':
                        styleSheet.body.stroke = extractColor(declaration)

                    elif declaration.name == 'stroke-width':
                        styleSheet.body.strokeWidth = extractDimensionalValue(declaration)

                    elif declaration.name == 'stroke-linecap':
                        styleSheet.body.strokeLinecap = extractEnum(declaration, StrokeLinecap)

            except ValueError as err:
                if len(err.args) and isinstance(err.args[0], tinycss2.ast.Declaration):
                    errorDeclaration: tinycss2.ast.Declaration = err.args[0]
                    errorMessage: str = err.args[1]
                    message = 'ERROR: {} (line {}, row {}): {}\n'.format(configFileName,
                                                                       errorDeclaration.source_line,
                                                                       errorDeclaration.source_column,
                                                                       errorMessage)
                    sys.stderr.write(message)
                    sys.exit(1)

        elif className in ('register', 'field'):
            for declaration in declarations:
                if declaration.name == 'fill':
                    getattr(styleSheet, className).fill = extractColor(declaration)

                elif declaration.name == 'stroke':
                    getattr(styleSheet, className).stroke = extractColor(declaration)

                elif declaration.name == 'stroke-width':
                    getattr(styleSheet, className).strokeWidth = extractDimensionalValue(declaration)

                elif declaration.name == 'stroke-linecap':
                    getattr(styleSheet, className).strokeLinecap = extractEnum(declaration, StrokeLinecap)

        elif className == 'field-name-line':
            for declaration in declarations:
                if declaration.name == 'stroke':
                    getattr(styleSheet, classNameMap[className]).stroke = extractColor(declaration)

                elif declaration.name == 'stroke-width':
                    getattr(styleSheet, classNameMap[className]).strokeWidth = extractDimensionalValue(declaration)

                elif declaration.name == 'stroke-linecap':
                    getattr(styleSheet, classNameMap[className]).strokeLinecap = extractEnum(declaration, StrokeLinecap)


        elif className in ('register-name', 'field-name', 'field-index'):
            for declaration in declarations:
                if declaration.name == 'font-family':
                    getattr(styleSheet, classNameMap[className]).fontFamily = extractFontFamily(declaration)

                elif declaration.name == 'font-size':
                    getattr(styleSheet, classNameMap[className]).fontSize = extractFontSize(declaration)

                elif declaration.name == 'font-style':
                    getattr(styleSheet, classNameMap[className]).fontStyle = extractEnum(declaration, FontStyle)

                elif declaration.name == 'font-weight':
                    getattr(styleSheet, classNameMap[className]).fontWeight = extractEnum(declaration, FontWeight)

                elif declaration.name == 'fill':
                    getattr(styleSheet, classNameMap[className]).fill = extractColor(declaration)


def cascadeStyles(styleSheet):
    ## Cascade styles in StyleSheet
    # body > register, field if property.value is None
    # body > register-name, field-name, field-index if property.value is None
    # cascade rect style

    obj = styleSheet.register
    if getattr(obj, 'fill') is None:
        setattr(obj, 'fill', styleSheet.body.fill)
    if getattr(obj, 'stroke') is None:
        setattr(obj, 'stroke', styleSheet.body.stroke)
    if getattr(obj, 'strokeWidth') is None:
        setattr(obj, 'strokeWidth', styleSheet.body.strokeWidth)
    if getattr(obj, 'strokeLinecap') is None:
        setattr(obj, 'strokeLinecap', styleSheet.body.strokeLinecap)

    obj = styleSheet.field
    if getattr(obj, 'fill') is None:
        setattr(obj, 'fill', styleSheet.register.fill)
    if getattr(obj, 'stroke') is None:
        setattr(obj, 'stroke', styleSheet.register.stroke)
    if getattr(obj, 'strokeWidth') is None:
        setattr(obj, 'strokeWidth', styleSheet.register.strokeWidth)
    if getattr(obj, 'strokeLinecap') is None:
        setattr(obj, 'strokeLinecap', styleSheet.register.strokeLinecap)

    obj = styleSheet.fieldNameLine
    if getattr(obj, 'stroke') is None:
        setattr(obj, 'stroke', styleSheet.field.stoke)
    if getattr(obj, 'strokeWidth') is None:
        setattr(obj, 'strokeWidth', styleSheet.field.strokeWidth)
    if getattr(obj, 'strokeLinecap') is None:
        setattr(obj, 'strokeLinecap', styleSheet.field.xstrokeLinecap)

    obj = styleSheet.registerName
    if getattr(obj, 'fontFamily') is None:
        setattr(obj, 'fontFamily', styleSheet.body.fontFamily)
    if getattr(obj, 'fontStyle') is None:
        setattr(obj, 'fontStyle', styleSheet.body.fontStyle)
    if getattr(obj, 'fontWeight') is None:
        setattr(obj, 'fontWeight', styleSheet.body.fontWeight)
    if getattr(obj, 'fontSize') is None:
        setattr(obj, 'fontSize', styleSheet.body.fontSize)
    # !!!: Because fill attribute used for both text and rect with different semantics, using rect.stroke value for text.fill
    if getattr(obj, 'fill') is None:
        setattr(obj, 'fill', styleSheet.body.stroke)

    for obj in [styleSheet.fieldName, styleSheet.fieldIndex]:
        if getattr(obj, 'fontFamily') is None:
            setattr(obj, 'fontFamily', styleSheet.registerName.fontFamily)
        if getattr(obj, 'fontStyle') is None:
            setattr(obj, 'fontStyle', styleSheet.registerName.fontStyle)
        if getattr(obj, 'fontWeight') is None:
            setattr(obj, 'fontWeight', styleSheet.registerName.fontWeight)
        if getattr(obj, 'fontSize') is None:
            setattr(obj, 'fontSize', styleSheet.registerName.fontSize)
        if getattr(obj, 'fill') is None:
            setattr(obj, 'fill', styleSheet.registerName.fill)


def extractEnum(declaration, enumClass):
    value = None
    tokens = list(
        filter(lambda x: isinstance(x, tinycss2.ast.IdentToken), declaration.value))
    if len(tokens) == 0:
        raise ValueError(declaration, '{} has no value.'.format(declaration.name))

    token = tokens[0]
    if token.value in [x.value for x in list(enumClass)]:
        value = token.lower_value
    else:
        legalValues = [x.value for x in list(enumClass)]
        message = '{} has a non-conformant value "{}". Legal values are {}'.format(declaration.name,
                                                                                   token.value,
                                                                                   legalValues)
        raise ValueError(declaration, message)

    return enumClass[token.value]

def extractDimensionalValue(declaration):
    value = None
    tokens = list(
        filter(lambda x: (isinstance(x, tinycss2.ast.IdentToken) or
                          isinstance(x, tinycss2.ast.DimensionToken) or
                          isinstance(x, tinycss2.ast.NumberToken)),
                          declaration.value))
    if len(tokens) == 0:
        raise ValueError(declaration, '{} has no value.'.format(declaration.name))

    token = tokens[0]
    if isinstance(token, tinycss2.ast.IdentToken):
        value = token.lower_value
    elif isinstance(token, tinycss2.ast.DimensionToken):
        value = '{}{}'.format(token.value, token.lower_unit)
    elif isinstance(token, tinycss2.ast.NumberToken):
        value = str(token.value)
    return value

def extractColor(declaration):
    tokens = list(
        filter(lambda x: (isinstance(x, tinycss2.ast.IdentToken) or
                          isinstance(x, tinycss2.ast.HashToken)),
               declaration.value))
    if len(tokens) == 0:
        raise ValueError(declaration, '{} has no value.'.format(declaration.name))

    token = tokens[0]
    if isinstance(token, tinycss2.ast.IdentToken):
        value = token.lower_value
    else:
        value = '#{}'.format(token.value)
    return value

def extractFontSize(declaration):
    tokens = list(
        filter(lambda x: (isinstance(x, tinycss2.ast.DimensionToken) or
                          isinstance(x, tinycss2.ast.NumberToken)),
                          declaration.value))
    if len(tokens) == 0:
        raise ValueError(declaration, '{} has no value.'.format(declaration.name))

    token = tokens[0]
    if isinstance(token, tinycss2.ast.DimensionToken):
        value = '{}{}'.format(token.value, token.unit)
    else:
        value = str(token.value)
    return value

def extractFontFamily(declaration):
    tokens = filter(lambda x: isinstance(x, tinycss2.ast.IdentToken), declaration.value)
    return [x.value for x in tokens]

def filterDeclarationErrors(configFileName, declarations):
    parseErrors = filter(lambda x: isinstance(x, tinycss2.ast.ParseError), declarations)
    for parseError in parseErrors:
        message = 'ERROR: {}: line {}, col {}: {}\nExiting.\n'.format(configFileName,
                                                                      parseError.source_line,
                                                                      parseError.source_column,
                                                                      parseError.message)

        sys.stderr.write(message)
        sys.exit(1)



##
# Copyright 2022 Charles Y. Choi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from regfmt import Constants
from regfmt import Endian
from regfmt import Field

DEFAULT_WIDTH = Constants.DEFAULT_WIDTH

class Register:
    def __init__(self, parent, config=None, width: int=DEFAULT_WIDTH, endian: Endian=Endian.bigByte):
        self.name: str = None
        self.width: int = width
        self.endian: Endian = endian
        self.fields: [Field] = None
        # parent is type TopLevel
        self.parent = parent
        if config:
            self.initFromConfig(config)

    def initFromConfig(self, config):
        self.name = config['name']
        self.width = config['width'] if 'width' in config else self.parent.width
        self.endian = config['endian'] if 'endian' in config else self.parent.endian

        if 'fields' in config:
            fields = []
            for fieldConfig in config['fields']:
                field = Field(config=fieldConfig)
                fields.append(field)

            self.fields = fields



##
# Copyright 2022 Charles Y. Choi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from regfmt import Constants
from regfmt import Register
from regfmt import Endian
from enum import Enum

DEFAULT_WIDTH = Constants.DEFAULT_WIDTH


class FieldNameAlign(Enum):
    center = "center"
    stairLeft = 'stair-left'


class RegisterLayout(Enum):
    tb = "tb"


class Layout:
    def __init__(self,
                 fieldNameAlign: FieldNameAlign = FieldNameAlign.center,
                 registerLayout: RegisterLayout = RegisterLayout.tb):
        self.fieldNameAlign: FieldNameAlign = fieldNameAlign
        self.registerLayout: RegisterLayout = registerLayout


class TopLevel:
    def __init__(self, config=None, width: int=DEFAULT_WIDTH, endian: Endian=Endian.bigByte):
        self.width: int = width
        self.endian: Endian = endian
        self.registers: [Register] = []
        self.layout = Layout()

        if config:
            self.width = config['width']
            self.endian = config['endian']

            if 'registers' in config:
                for registerConfig in config['registers']:
                    self.registers.append(Register(self, registerConfig))

            if 'layout' in config:
                if 'field-name-align' in config['layout']:
                    self.layout.fieldNameAlign = FieldNameAlign(config['layout']['field-name-align'])

                if 'register-layout' in config['layout']:
                    self.layout.registerLayout = RegisterLayout(config['layout']['register-layout'])



##
# Copyright 2022 Charles Y. Choi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from jsonschema import validate, exceptions
from yaml import Loader, Dumper, load, safe_load
from yaml.scanner import ScannerError
import sys

INPUT_SCHEMA_YAML="""
type: object
properties: 
  width:
     type: integer
     minimum: 1

  endian:
     $ref: "#/$defs/endian"

  registers:
     type: array
     items:
       $ref: "#/$defs/register"

  layout:
    type: object
    properties:
      field-name-align:
        enum:
          - center
          - stair-left

required:
  - width
  - registers
  - endian

$defs:
  endian:
    enum:
      - littleByte
      - bigByte
      - littleBit
      - bigBit

  field:
    type: object
    properties:
      name:
        type: 
          - string
          - "null"
      width:
        type: integer
    required:
      - width

  register:
    type: object
    properties:
      name:
        type: 
          - string
          - "null"
      width:
        type: integer

      endian:
        $ref: "#/$defs/endian"

      fields:
        type: array
        items:
          $ref: "#/$defs/field"

    required:
      - fields
"""


class InputLoadAndValidate:
    def __init__(self, parsedArguments):
        self.parsedArguments = parsedArguments

    def loadAndValidate(self):
        inputYAML = self.loadInput(self.parsedArguments)
        schemaYAML = self.loadInputSchema(INPUT_SCHEMA_YAML)

        try:
            result = validate(inputYAML, schemaYAML)
            if result is not None:
                # this should never happen
                sys.stderr.write('ERROR: unexpected result from validate()\n')
                sys.exit(1)

        except exceptions.ValidationError as err:
            # TODO: implement YAML ValidationError handling
            message = 'ERROR: YAML file "{}" in path "{}": {}'
            sys.stderr.write(message.format(self.parsedArguments.input,
                                            err.json_path,
                                            err.message))
            sys.exit(1)

            # print(err.json_path)
            # print(err.message)
            # print(err.path)
            # print(err.relative_path)
            # print(err.absolute_path)
            # print(err.context)
            # print(err.cause)
            # print(err.instance)
            # print(err.validator)
            # print(err.schema_path)
            # print(dir(err))

        return inputYAML
        
    def loadInput(self, parsedArguments):
        with open(parsedArguments.input, 'r') as infile:
            try:
                yamlConfig = load(infile, Loader=Loader)
            except ScannerError as err:
                sys.stderr.write('ERROR: {0}\n'.format(err))
                sys.exit(1)

        return yamlConfig

    def loadInputSchema(self, schemaString):
        try:
            yamlConfig = safe_load(schemaString)
        except ScannerError as err:
            sys.stderr.write('ERROR: {0}\n'.format(err))
            sys.exit(1)

        return yamlConfig



from regfmt.svggeometry import *
from regfmt import Register


def writeFieldNameCenterSVG(registerDB, topGroup, styleSheet, font, registerSpacing=5):
    bitFieldSize = getBitFieldSize(font=font)

    registers: [Register] = registerDB.registers
    parentOrigin = Point()
    parentSize = Size(height=(bitFieldSize.height + registerSpacing) * len(registers))

    displacementY = 0

    for register in registers:
        registerGroup = Group()
        topGroup.append(registerGroup)

        registerHeight = bitFieldSize.height
        registerWidth = bitFieldSize.width * register.width

        if register.name:
            registerNameFrame = getTextFrame(text=register.name, font=styleSheet.registerName.truetype())

            x = registerWidth + 10
            y = registerNameFrame.size.height + (registerHeight/2.0) - (registerNameFrame.size.height/2.0)

            registerGroup.append(Text(text=register.name,
                                      x=x,
                                      y=y,
                                      style=styleSheet.registerName))

            newParentWidth = x + registerNameFrame.size.width
            if parentSize.width < newParentWidth:
                parentSize.width = newParentWidth

        displacementX = 0

        for field in register.fields:
            fieldWidth = field.width * bitFieldSize.width
            fieldHeight = registerHeight

            T = translateTransform(displacementX, 0.0)
            V = vector2D(0.0, 0.0)
            transformedV = matrixMult(T, V)
            fieldX, fieldY = coordinateFromVector2D(transformedV)

            registerGroup.append(Rect(x=fieldX,
                                      y=fieldY,
                                      width=fieldWidth,
                                      height=fieldHeight,
                                      style=styleSheet.field))

            if field.name:
                fieldNameFrame = getTextFrame(text=field.name, font=styleSheet.fieldName.truetype())

                fieldNameX = (fieldWidth / 2.0) + displacementX
                fieldNameY = fieldNameFrame.origin.y + (fieldHeight / 2.0) - (fieldNameFrame.size.height / 2.0)

                registerGroup.append(Text(text=field.name,
                                          x=fieldNameX,
                                          y=fieldNameY,
                                          textAnchor='middle',
                                          style=styleSheet.fieldName))

            if field.width > 1:
                registerGroup.append(Text(str(field.leftIndex),
                                          x=fieldX + 3,
                                          y=fieldY + fieldHeight - 3,
                                          textAnchor='start',
                                          style=styleSheet.fieldIndex
                                          ))

                registerGroup.append(Text(str(field.rightIndex),
                                          x=fieldX + fieldWidth - 3,
                                          y=fieldY + fieldHeight - 3,
                                          textAnchor='end',
                                          style=styleSheet.fieldIndex
                                          ))
            else:
                registerGroup.append(Text(str(field.leftIndex),
                                          x=fieldX + (fieldWidth / 2.0),
                                          y=fieldY + fieldHeight - 3,
                                          textAnchor='middle',
                                          style=styleSheet.fieldIndex
                                          ))

            displacementX += fieldWidth

    for group in topGroup:
        group.translate(0, displacementY)
        displacementY += (bitFieldSize.height + registerSpacing)

    return (parentOrigin, parentSize)




##
# Copyright 2022 Charles Y. Choi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from os import uname

VERSION = '0.1.10'
BASE_FONT_NAME = 'Helvetica'

unameObj = uname()
if unameObj.sysname == 'Linux':
    BASE_FONT_NAME = 'FreeSans'
elif unameObj.sysname == 'Windows':
    BASE_FONT_NAME = 'Arial'

from .InputLoadAndValidate import InputLoadAndValidate
from .Constants import Constants
from .Endian import Endian
from .Field import Field
from .Register import Register
from .TopLevel import FieldNameAlign, RegisterLayout, Layout, TopLevel
from .DRCChecker import DRCChecker
from .cssstyles import *
from .svggeometry import *
from .svgwriter import SVGWriter
from .centeralignlayout import *
from .stairleftlayout import *
from .CommandLineParser import CommandLineParser
from .RegisterFormat import RegisterFormat


def main():
    app = RegisterFormat(CommandLineParser().run())
    app.run()



##
# Copyright 2022 Charles Y. Choi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from regfmt import Register
from regfmt import Field
from regfmt import TopLevel
from regfmt import Endian
from operator import add
from functools import reduce
import sys

class DRCChecker:
    def __init__(self):
        self.foo = None
        self.errors = []


    def check(self, registerDB: TopLevel):
        # TODO: this should be a throwing method
        self.checkFieldWidths(registerDB)

    def checkFieldWidths(self, registerDB: TopLevel):
        registers: [Register] = registerDB.registers

        for register in registers:
            fields: [Field] = register.fields
            widths = [x.width for x in fields]
            sum = reduce(add, widths)
            if sum != register.width:
                message = 'DRC Violation: sum total of widths of fields ({0}) must equal the register width ({1}) for register {2}\n'.format(sum, register.width, register.name)
                self.errors.append(message)

        for message in self.errors:
            sys.stderr.write(message)

        return self.errors

    # TODO: check that field names do not repeat
    # TODO: check that fields are byte-sized for bigByte and littleByte

    def subIndexFields(self, registerDB: TopLevel):
        registers: [Register] = registerDB.registers

        for register in registers:
            endian = register.endian
            if endian in (Endian.bigBit.value, Endian.bigByte.value):
                count = register.width
            elif endian in (Endian.littleBit.value, Endian.littleByte.value):
                count = -1
            fields: [Field] = register.fields

            for field in fields:
                if endian in (Endian.bigBit.value, Endian.bigByte.value):
                    count -= 1
                    field.leftIndex = count
                    count -= (field.width - 1)
                    field.rightIndex = count
                    #print('{0}:{1} {2}'.format(field.leftIndex, field.rightIndex, field.name))

                elif endian == Endian.littleBit.value:
                    count += 1
                    field.leftIndex = count
                    count += (field.width - 1)
                    field.rightIndex = count
                    #print('{0}:{1} {2}'.format(field.leftIndex, field.rightIndex, field.name))

                elif endian == Endian.littleByte.value:
                    count += 1
                    field.rightIndex = count
                    count += (field.width - 1)
                    field.leftIndex = count
                    #print('{0}:{1} {2}'.format(field.leftIndex, field.rightIndex, field.name))






from regfmt.svggeometry import *
from regfmt import Register

def writeStairLeftSVG(registerDB, topGroup, styleSheet, font, registerSpacing=5):
    bitFieldSize = getBitFieldSize(font=font)

    registers: [Register] = registerDB.registers
    parentOrigin = Point()
    parentSize = Size(height=(bitFieldSize.height + registerSpacing) * len(registers))

    maxFieldDisplacementYs = []
    displacementY = 0
    for register in registers:
        registerGroup = Group()
        topGroup.append(registerGroup)

        registerHeight = bitFieldSize.height
        registerWidth = bitFieldSize.width * register.width

        if register.name:
            registerNameFrame = getTextFrame(text=register.name, font=styleSheet.registerName.truetype())

            x = registerWidth + 10
            y = registerNameFrame.size.height + (registerHeight/2.0) - (registerNameFrame.size.height/2.0)

            registerGroup.append(Text(text=register.name,
                                      x=x,
                                      y=y,
                                      style=styleSheet.registerName))

            newParentWidth = x + registerNameFrame.size.width
            if parentSize.width < newParentWidth:
                parentSize.width = newParentWidth

        displacementX = 0

        fieldNames = list(filter(lambda x: x is not None, [field.name for field in register.fields]))
        fieldCount = len(fieldNames)

        fieldFrames = tuple(map(lambda x: getTextFrame(text=x, font=styleSheet.fieldName.truetype()), fieldNames))
        fieldHeights = tuple(map(lambda x: x.size.height, fieldFrames))
        maxFieldHeight = max(fieldHeights)

        fieldDisplacementY = (maxFieldHeight + 10.0) * fieldCount
        maxFieldDisplacementYs.append(fieldDisplacementY)

        counter = 0
        for field in register.fields:
            fieldWidth = field.width * bitFieldSize.width
            fieldHeight = registerHeight

            T = translateTransform(displacementX, 0.0)
            V = vector2D(0.0, 0.0)
            transformedV = matrixMult(T, V)
            fieldX, fieldY = coordinateFromVector2D(transformedV)

            registerGroup.append(Rect(x=fieldX,
                                      y=fieldY,
                                      width=fieldWidth,
                                      height=fieldHeight,
                                      style=styleSheet.field))

            if field.width > 1:
                registerGroup.append(Text(str(field.leftIndex),
                                          x=fieldX + 3,
                                          y=fieldY + fieldHeight - 3,
                                          textAnchor='start',
                                          style=styleSheet.fieldIndex
                                          ))

                registerGroup.append(Text(str(field.rightIndex),
                                          x=fieldX + fieldWidth - 3,
                                          y=fieldY + fieldHeight - 3,
                                          textAnchor='end',
                                          style=styleSheet.fieldIndex
                                          ))
            else:
                registerGroup.append(Text(str(field.leftIndex),
                                          x=fieldX + (fieldWidth / 2.0),
                                          y=fieldY + fieldHeight - 3,
                                          textAnchor='middle',
                                          style=styleSheet.fieldIndex
                                          ))

            if field.name:
                fieldNameFrame = fieldFrames[counter]

                fieldNameX = (fieldWidth / 2.0) + displacementX + 3.0
                fieldNameY = fieldHeight + fieldNameFrame.origin.y + fieldDisplacementY + 10.0

                newParentWidth = fieldNameX + fieldNameFrame.size.width
                if parentSize.width < newParentWidth:
                    parentSize.width = newParentWidth

                registerGroup.append(Text(text=field.name,
                                          x=fieldNameX,
                                          y=fieldNameY,
                                          textAnchor='left',
                                          style=styleSheet.fieldName))

                registerGroup.append(Line(x1=displacementX + (fieldWidth / 2.0),
                                          y1=registerHeight,
                                          x2=displacementX + (fieldWidth / 2.0 ),
                                          y2=fieldNameY,
                                          style=styleSheet.fieldNameLine))

                fieldDisplacementY -= (maxFieldHeight + 10)

                if parentSize.height < fieldNameY:
                    parentSize.height = fieldNameY

                counter += 1

            displacementX += fieldWidth

    counter2 = 0
    for group in topGroup:
        group.translate(0, displacementY)
        displacementY += (bitFieldSize.height * 2.0 + registerSpacing + maxFieldDisplacementYs[counter2])
        if parentSize.height < displacementY:
            parentSize.height = displacementY
        counter2 += 1

    return (parentOrigin, parentSize)

